; Harald Peters, 01.01.11
; FuseE F9h -> FCh: Bootloader 512Byte an D0, D1
; FuseH DFh -> D4h: EE nicht lschen, BOD = 4.3V
; FuseL 62h -> FFh: 1:1, ext. Quarz
.include "m88pdef.inc"
;.include "m8def.inc"

; use 14.7456MHz if you need the UART
;.equ    F_CPU       = 16000000
;.equ   F_CPU       = 14745600
.equ   F_CPU        = 22118400

.equ    BAUD        = 115200

.equ UBRR_VAL   = ((F_CPU+BAUD*8)/(BAUD*16)-1)  ; clever runden
.equ BAUD_REAL  = (F_CPU/(16*(UBRR_VAL+1)))     ; Reale Baudrate
.equ BAUD_ERROR = ((BAUD_REAL*1000)/BAUD-1000)  ; Fehler in Promille

.if ((BAUD_ERROR>10) || (BAUD_ERROR<-10))       ; max. +/-10 Promille Fehler
.warning "baudrate error > 1%, UART disabled"
.endif
; source: http://www.mikrocontroller.net/articles/AVR-Tutorial:_UART

; http://wwwlehre.dhbw-stuttgart.de/~schulte/digiisdn.htm
; http://www.virtualuniversity.ch/telekom/telefon/7.html
; http://www.rhyshaden.com/isdn.htm

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; r0/1: SPM/LPM

.def    t0val       = r2
.def    lastpol     = r3

; immediates
.def    temp        = r16
.def    temp1       = r17
.def    temp2       = r18

.def    state       = r19
.def    bitcnt      = r20
.def    dByte       = r21
.def    dbitcnt     = r22
.def    einscnt     = r23
.def    dzust       = r24
.def    bytecnt     = r25

; X: 26/27
; Y: 28/29
; Z: 30/31
; Zustand in state gespeichert
.equ    ST_NOSYNC   = 0
.equ    ST_F        = 1
.equ    ST_L        = 2
.equ    ST_DATA     = 3

; 16MHz/192000 = 83.333
.equ    T0_SVAL     = (F_CPU/192000)

.dseg
TxPuffer:	.byte	256			; 

.cseg
.org 0
    rjmp RESET
.org INT0addr
    rjmp INT0_ISR
.org INT1addr
    rjmp INT1_ISR
.org OVF0addr
    rjmp OVF0_ISR


OVF0_ISR:
    ; prepare T0 for next sample
    ; adjust t0val to compensate for
    ; the IRQ->ISR delay (current val of T0CNT)
    ; just add delay to t0val
    in temp2, TCNT0
    add temp2, t0val
    out TCNT0, temp2
    
    rjmp SAMPLE_S0


INT0_ISR:
INT1_ISR:
    ; prepare T0 for next sample
    ; adjust t0val to compensate for
    ; the IRQ->ISR delay (current val of T0CNT)
    ; just add delay to t0val
    out TCNT0, t0val
    
    ; reset T0OVF flag
    nop
    nop
    nop
    ldi temp, (1<<TOV0)
    out TIFR0, temp
    
    ; enable T0 int
    ldi temp, (1<<TOIE0)
    sts TIMSK0, temp
    
SAMPLE_S0:
    in temp, PIND
    andi temp, $0C
    
    ; clear interrupt flags
    ldi temp2, (1<<INTF1)|(1<<INTF0)
    out EIFR, temp2
    
    ; "11" -> '1', else '0'
    cpi temp, $0C
    brne ZERO_BIT
    
    ; '1' databit
    ldi temp, 0x80    ; Bit7 = 1 wegen einfacherer Ser.-/Par.- Wandlung
    
    cpi state, ST_L
    brsh RCV_DATABIT

SYNC_LOST:
    clr temp
    sts TIMSK0, temp
    
SYNC_LOST2:
    clr state
    reti
    
ZERO_BIT:
    ; same polarity as last sample?
    ; (code violation - sync candidate)
    cp temp, lastpol
    breq CODE_VIOLATION
    
    ; no code violation
    
    ; save polarity for next sample
    mov lastpol, temp
    
    ; inverse AMI: +-diff is '0' databit
    clr temp
    
    ; normal databit or
    ; F -> L state transition?
    cpi state, ST_L
    brsh RCV_DATABIT
    
    cpi state, ST_F
    brne SYNC_LOST2
    
    ; F -> L state transition

    ; L bit received
    ; -> start data reception
    ldi state, ST_L
    
    ; reset bitcnt (L-bit is #2)
    ldi bitcnt, 2
    reti
    
    
CODE_VIOLATION: ; same polarity as last sample
    
    ; there's a 2nd code violation for
    ; 1st zero databit after L bit
    ; don't resync in this case
    cpi state, ST_L
    breq SYNC_IGNORE
    
    ; 1st code violation -> resync
    ldi state, ST_F    
    reti
    
    
SYNC_IGNORE:
    ; 2nd code violation
    ; treat further code violations as sync
    ldi state, ST_DATA
    
    ; inverse AMI: +-diff is '0' databit
    clr temp
    
    
RCV_DATABIT: 
    cpi bitcnt, 47   ; check for end of frame, alle 250us (4KHz)
    brsh RCV_RESET

    inc bitcnt

    ; D-Bits
    cpi bitcnt, 12
    breq HANDLE_DBit
    cpi bitcnt, 25
    breq HANDLE_DBit
    cpi bitcnt, 36
    breq HANDLE_DBit
    cpi bitcnt, 47
    breq HANDLE_DBit   
    reti
    

RCV_RESET:
    ldi state, ST_NOSYNC
    reti


HANDLE_DBIT:
    cpi  dzust, 0
    breq Zust0      ; Zust=0: warten auf Bit=0
    cpi  dzust, 1
    breq Zust1      ; Zust=1: warten auf Startzeichen 0xE7
    rjmp Zust2      ; Zust=2,3; Datenbytes mit evtl. Stopfbits

Zust0:              ; warten auf 0-Bit
    sbrc temp, 7    ; wenn temp==0, dann erstes Bit, nur Bit7
    reti
; erste 0 erkannt
    ldi   dzust, 1
    clr   dbitcnt
    clr   dByte
    rcall SerPar
    reti

Zust1:              ; warten auf 7Eh-Zeichen
    rcall SerPar
    cpi   dbitcnt, 8
    breq  BFert
    reti

BFert:
    cpi  dByte, 0x7E
    breq Zust1a     ; Startzeichen erkannt
    clr  dzust
    rjmp S1         ; weiter warten
Zust1a:
    inc  dzust      ; Zust=2
    rjmp ByteFert

; ab hier Datenbytes
Zust2:
    sbrc temp, 7    ; Einsen zhlen
    rjmp Bit1       ; wird bersprungen, wenn Bit=0
; Bit=0
    cpi  einscnt, 5 ;
    breq EntferneStopfBit
    clr  einscnt
    rjmp nBit

EntferneStopfBit:   ; Bit wird nicht mitgezhlt
    clr  einscnt    ; erstes Nullbit setzt einscnt auf Null
    reti

Bit1: 
    inc  einscnt
    cpi  einscnt, 6 ;
    breq Zust2a     ; 6 Einsen -> Rahmen fertig
nBit:
    rcall SerPar
    cpi  dbitcnt, 8
    breq ByteFert   ; 8 Bits eingelesen
    reti

Zust2a:             ; 6 Eisen gezhlt
    inc  dzust      ; fertig: Zust=3
    rjmp nBit

SerPar:
    lsr  dByte
    or   dByte,temp
    inc  dbitcnt
    ret


ByteFert:            ; Daten ins RAM
	st   X+, dByte
    inc  bytecnt
    cpi  dzust, 3    ; wenn dzust==3, dann Rahmen gelesen
    breq Zust2Ferta
S1:
    clr  dbitcnt
    clr  dByte
    reti

Zust2Ferta:
    clr  dzust

    cpi  bytecnt, 10     ; hier nur Ausgabe, wenn >= 10 Bytes
    brsh SendRAM         ; nur lange Datenfolgen ausgeben
SendRAMFert:
	ldi  XL, low(TxPuffer)  ; RAM-Zeiger auf Ursprung
	ldi  XH, high(TxPuffer)
    clr  bytecnt
    rjmp S1


SendRAM:
	ldi XL, low(TxPuffer)
	ldi XH, high(TxPuffer)
Sr:
    ld    dByte, X+
    rcall SendASCII
    ldi   dByte, ' '
    rcall SendZeich
    dec   bytecnt
    brne  Sr

    ldi   dByte, 13
    rcall SendZeich
    ldi   dByte, 10
    rcall SendZeich
    rjmp  SendRAMFert


; Prozedur SendASCII: dByte in 2 ASCII-Zeichen
SendASCII:
    mov   temp1, dByte
    swap  dByte
    rcall SendHByte
    mov   dByte, temp1
    rcall SendHByte
    ret

SendHByte:
    andi  dByte, 0x0F
    cpi   dByte, 10        ; 0..9 -> 30h..39h
    brsh  SA1
    ori   dByte, 0x30
    rcall SendZeich
    ret
SA1:
    subi  dByte, 201       ; A..F -> 41h..46h
    rcall SendZeich
    ret


SendZeich:
Warte:	
    lds  temp2,UCSR0A	
    sbrs temp2,UDRE0  	; skip if UDRE=1 (USART Data Register Empty)
	rjmp Warte
	sts	 UDR0,dByte
    ret
;--------------------------------   
RESET:
    ; setup stack 1st
    ldi temp, high(RAMEND)
    out SPH, temp
    ldi temp, low(RAMEND)
    out SPL, temp
    
    ; enable T0 with F_CPU
    ; for sampling the S0 bus
    ldi temp, (1<<CS00)
    out TCCR0B, temp
    
    
    ; enable output for LEDs
    ldi temp, (1<<PB0)|(1<<PB1)|(1<<PB2)|(1<<PB3)
    out DDRB, temp
    ldi temp, (1<<PB0) ; gn LED ein
    out PORTB, temp
    
    ; INT0/1 falling edges
    ldi temp, (1<<ISC01)|(1<<ISC11)
    sts EICRA, temp
    ldi temp, (1<<INT0)|(1<<INT1)
    out EIMSK, temp
    
    clr state
    clr lastpol
    clr dbitcnt
    clr dzust
	clr bytecnt
	ldi XL, low(TxPuffer)
	ldi XH, high(TxPuffer)
    
    ; prepare t0vals for S0 sampling
    ; single databit distance
    ldi temp, low(256 - T0_SVAL)    ; T0_SVAL = 83
    mov t0val, temp
    
.if ((BAUD_ERROR<=10) && (BAUD_ERROR>=-10))
; UART init
    ldi   temp2, high(UBRR_VAL)
    ldi   temp, low(UBRR_VAL)
    sts   UBRR0H, temp2
    sts   UBRR0L, temp
    ldi   temp, (1<<TXEN0)
    sts   UCSR0B, temp
; Text ber Uart ausgeben
	ldi	  ZH,high(2*AText);	; load high part of byte address
	ldi	  ZL,low(2*AText);	; load low part of byte address
LadeTxt:
	lpm				; load memory byte into r0
	tst   r0		; Endezeichen = 0
	breq  EndeText	; fertig
    mov   dByte, r0
    rcall SendZeich
	adiw  ZL,1		; nchstes Byte
	rjmp  LadeTxt
EndeText:
.endif

    sei
    
MAIN:
    rjmp MAIN

AText:
.db	"Harald Peters, V2.00, 01.01.2011",	13,10,0,0

